############################################################################
#      ____  _____________  __    __  __ _           _____ ___   _         #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)   #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \       #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\      #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B      #
#                                                                          #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.           #
#                    2025 Advanced Micro Devices, Inc. All rights reserved #
#                  SPDX-License-Identifier: BSD-3-Clause                   #
#                                                                          #
############################################################################

from frugy.types import *
from frugy.multirecords import ipmi_multirecord, MultirecordEntry
from frugy.fru_registry import FruRecordType, rec_register, rec_lookup_by_id


# IPMI standard multirecords


@ipmi_multirecord(0xd2)
class OemXilinxEntry(MultirecordEntry):
    ''' Xilinx proprietary, not published '''

    _xilinx_identifier = 0x10da

    def _payload_prologue(self):
        return self._xilinx_identifier.to_bytes(3, 'little') + self._record_id.to_bytes(length=1, byteorder='little')

    @classmethod
    def from_payload(cls, payload):
        xilinx_id, payload = payload[:3], payload[3:]
        xilinx_id = int.from_bytes(xilinx_id, 'little')
        rec_id, payload = payload[0], payload[1:]


        if xilinx_id != cls._xilinx_identifier:
            raise ValueError(
                f"FMC identifier mismatch: expected 0x{cls._xilinx_identifier:06x}, received 0x{xilinx_id:06x} ({xilinx_id})")
        try:
            cls_inst = rec_lookup_by_id(
                FruRecordType.xilinx_multirecord, rec_id)()
        except KeyError:
            raise RuntimeError(f"Unknown Xilinx entry 0x{rec_id:02x}")

        cls_inst._deserialize(payload)
        cls_inst._record_id = rec_id
        return cls_inst


@ipmi_multirecord(0xd3)
class OemXilinxD3Entry(MultirecordEntry):
    ''' Xilinx proprietary, not published '''

    _xilinx_identifier = 0x10da

    def _payload_prologue(self):
        return self._xilinx_identifier.to_bytes(3, 'little')

    @classmethod
    def from_payload(cls, payload):
        xilinx_id, payload = payload[:3], payload[3:]
        rec_id = 0xd3

        try:
            cls_inst = rec_lookup_by_id(
                FruRecordType.xilinx_multirecord, rec_id)()
        except KeyError:
            raise RuntimeError(f"Unknown Xilinx entry 0x{rec_id:02x}")

        cls_inst._deserialize(payload)
        cls_inst._record_id = rec_id
        return cls_inst

def xilinx_multirecord(rec_id):
    def register_and_set_id(cls):
        cls._record_id = rec_id
        rec_register(cls, FruRecordType.xilinx_multirecord, rec_id)
        return cls
    return register_and_set_id

@xilinx_multirecord(0x31)
class DutXilinxMac(OemXilinxEntry):
    _schema = [
        ('mac0', MacField),
        ('mac1', MacField, None),
        ('mac2', MacField, None),
        ('mac3', MacField, None),
    ]

    def to_dict(self):
        result = super().to_dict()
        new_result = {}
        for key, value in result.items():
            if (value != ''):
                new_result[key] = value
        return new_result



@xilinx_multirecord(0x11)
class SysCtrlXilinxMac(OemXilinxEntry):
    _schema = [
        ('mac0', MacField),
        ('mac1', MacField, None),
        ('mac2', MacField, None),
        ('mac3', MacField, None),
    ]

    def to_dict(self):
        result = super().to_dict()
        new_result = {}
        for key, value in result.items():
            if (value != ''):
                new_result[key] = value
        return new_result

@xilinx_multirecord(0xd3)
class XilinxOemD3(OemXilinxD3Entry):
    _schema = [
        ('data', FixedStringField, 80, {'default': ''}),
    ]




###########################################################################
#      ____  _____________  __    __  __ _           _____ ___   _        #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)  #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \      #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\     #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B     #
#                                                                         #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.          #
#                  SPDX-License-Identifier: BSD-3-Clause                  #
#                                                                         #
###########################################################################

from frugy.types import FruAreaVersioned, FruAreaSized, FixedField, StringField, StringField, bin2hex_helper, CustomStringArray
from frugy.fru_registry import FruRecordType, rec_register
from datetime import datetime, timedelta
import logging
import frugy

def ipmi_area(cls):
    rec_register(cls, FruRecordType.ipmi_area)
    return cls


_default_language_code = 0  # Use English and UTF-8 as default encoding


@ipmi_area
class CommonHeader(FruAreaVersioned):
    ''' Platform Management FRU Information Storage Definition, Table 8-1 '''

    _schema = [
        ('internal_use_offs', FixedField, 'u8', {'default': 0}),
        ('chassis_info_offs', FixedField, 'u8', {'default': 0}),
        ('board_info_offs', FixedField, 'u8', {'default': 0}),
        ('product_info_offs', FixedField, 'u8', {'default': 0}),
        ('multirecord_offs', FixedField, 'u8', {'default': 0}),
    ]

    def __getitem__(self, key):
        return self._get(key) * 8

    def __setitem__(self, key, value):
        if value % 8 != 0:
            raise RuntimeError("Offset not aligned to 64 bit")
        self._set(key, value // 8)

    def reset(self):
        for item in self._dict:
            self._set(item, 0)


@ipmi_area
class ChassisInfo(FruAreaSized):
    ''' Platform Management FRU Information Storage Definition, Table 10-1 '''

    _schema = [
        ('type', FixedField, 'u8'),
        ('part_number', StringField),
        ('serial_number', StringField),
        ('custom_info_fields', CustomStringArray),
    ]


@ipmi_area
class BoardInfo(FruAreaSized):
    ''' Platform Management FRU Information Storage Definition, Table 11-1 '''

    _schema = [
        ('language_code', FixedField, 'u8', {
         'default': _default_language_code}),
        ('mfg_date_time', FixedField, 'u24', {'default': 0}),
        ('manufacturer', StringField),
        ('product_name', StringField),
        ('serial_number', StringField),
        ('part_number', StringField),
        ('fru_file_id', StringField),
        ('custom_info_fields', CustomStringArray),
    ]

    # With a Y2K27 rollover baked into the specification, we have to deal with
    # - 24bit overflow when converting yaml to bin
    # - uncertainty of original 25th bit when converting bin to yaml
    # see also: https://github.com/MicroTCA-Tech-Lab/frugy/issues/11
    _time_ref = datetime(1996, 1, 1)
    _time_rollover_limit = 2**24

    def _timestamp_to_minutes(self, timestamp):
        td = timestamp - self._time_ref
        return td.seconds // 60 + td.days * (60*24)

    def _timestamp_from_minutes(self, minutes):
        return BoardInfo._time_ref + timedelta(minutes=minutes)

    def _handle_y2k27_rollover_yaml2bin(self, minutes):
        # Truncate the timestamp to 24 bits
        # If the remainder is nonzero, log a warning.
        minutes_truncated = minutes % BoardInfo._time_rollover_limit
        if minutes != minutes_truncated:
            min_str = self._timestamp_from_minutes(minutes).isoformat()
            min_trunc_str = self._timestamp_from_minutes(minutes_truncated).isoformat()
            logging.warning(f'Timestamp {min_str} truncated; may be interpreted by FRU parsers as {min_trunc_str}')
        return minutes_truncated

    def _handle_y2k27_rollover_bin2yaml(self, minutes, now):
        # Apply heuristic to possibly truncated timestamp:
        # If it is "too old" relative to the system time, add rollover timedelta and log a warning.
        # We draw the line at 31 years into the past (not the full 2**24 minutes).
        # This will allow a margin of a couple months in case the system time is incorrect
        # or the board is dated a bit into the future.
        if now - self._timestamp_from_minutes(minutes) > timedelta(days=31*365):
            minutes_ext = minutes + self._time_rollover_limit
            min_str = self._timestamp_from_minutes(minutes).isoformat()
            min_ext_str = self._timestamp_from_minutes(minutes_ext).isoformat()
            warn_str = f'BoardInfo.mfg_date_time: possible Y2K27 rollover detected; timestamp bumped to {min_ext_str} - original timestamp was {min_str}'
            logging.warning(warn_str)
            frugy.fru.import_log(warn_str)
            return minutes_ext
        return minutes

    def _set_mfg_date_time(self, timestamp):
        if timestamp is not None:
            if type(timestamp) == str:
                timestamp = datetime.fromisoformat(timestamp)
            minutes = self._timestamp_to_minutes(timestamp)
            minutes = self._handle_y2k27_rollover_yaml2bin(minutes)
            self._set('mfg_date_time', minutes)
        else:
            self._set('mfg_date_time', 0)

    def _get_mfg_date_time(self):
        minutes = self._get('mfg_date_time')
        if minutes != 0:
            minutes = self._handle_y2k27_rollover_bin2yaml(minutes, datetime.now())
            return self._timestamp_from_minutes(minutes)
        else:
            return None


@ipmi_area
class ProductInfo(FruAreaSized):
    ''' Platform Management FRU Information Storage Definition, Table 12-1 '''

    _schema = [
        ('language_code', FixedField, 'u8', {
         'default': _default_language_code}),
        ('manufacturer', StringField),
        ('product_name', StringField),
        ('part_number', StringField),
        ('version', StringField),
        ('serial_number', StringField),
        ('asset_tag', StringField),
        ('fru_file_id', StringField),
        ('custom_info_fields', CustomStringArray),
    ]



###########################################################################
#      ____  _____________  __    __  __ _           _____ ___   _        #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)  #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \      #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\     #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B     #
#                                                                         #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.          #
#                  SPDX-License-Identifier: BSD-3-Clause                  #
#                                                                         #
###########################################################################

from frugy.types import FixedField, BytearrayField, ser_6bit, deser_6bit
from frugy.multirecords import ipmi_multirecord, MultirecordEntry
from frugy.fru_registry import FruRecordType, rec_register, rec_lookup_by_id

import bitstruct
from bidict import bidict


@ipmi_multirecord(0xfa)
class FmcEntry(MultirecordEntry):
    ''' ANSI/VITA 57.1 FMC Standard '''
    ''' Superclass of all ANSI/VITA FMC OEM multirecords '''

    _fmc_identifier = 0x12a2

    def _payload_prologue(self):
        return self._fmc_identifier.to_bytes(3, 'little') + self._record_id.to_bytes(length=1, byteorder='little')

    @classmethod
    def from_payload(cls, payload):
        if not MultirecordEntry.opalkelly_workaround_enabled:
            # Opal Kelly FMC records seem to skip the manufacturer ID ...
            fmc_id, payload = payload[:3], payload[3:]
            rec_id, payload = payload[0], payload[1:]
            fmc_id = int.from_bytes(fmc_id, 'little')

            if fmc_id != cls._fmc_identifier:
                raise ValueError(
                    f"FMC identifier mismatch: expected 0x{cls._fmc_identifier:06x}, received 0x{fmc_id:06x} ({fmc_id})")
        else:
            # Opal Kelly FMC records seem to have the rec_id as _last_ instead of first byte
            rec_id, payload = payload[-1], payload[:-1]

        try:
            cls_inst = rec_lookup_by_id(
                FruRecordType.fmc_multirecord, rec_id)()
        except KeyError:
            raise RuntimeError(f"Unknown FMC entry 0x{rec_id:02x}")

        cls_inst._deserialize(payload)
        cls_inst._record_id = rec_id
        return cls_inst


# FMC multirecords

def fmc_multirecord(rec_id):
    def register_and_set_id(cls):
        cls._record_id = rec_id
        rec_register(cls, FruRecordType.fmc_multirecord, rec_id)
        return cls
    return register_and_set_id


def fmc_secondary(cls):
    rec_register(cls, FruRecordType.fmc_secondary)
    return cls


@fmc_multirecord(0x00)
class FmcMainDefinition(FmcEntry):
    ''' ANSI/VITA 57.1 FMC Standard, Table 7 '''

    _schema = [
        ('module_size', FixedField, 'u2', {'constants': {
            'single_width': 0b00,
            'double_width': 0b01
        }}),
        ('p1_connector_size', FixedField, 'u2', {'constants': {
            'lpc': 0b00,
            'hpc': 0b01
        }}),
        ('p2_connector_size', FixedField, 'u2', {'constants': {
            'lpc': 0b00,
            'hpc': 0b01,
            'not_fitted': 0b11,
        }}),
        ('clock_direction', FixedField, 'u1', {'constants': {
            'm2c': 0b0,
            'c2m': 0b1,
        }}),
        ('_reserved', FixedField, 'u1', {'default': 0}),
        ('p1_a_num_signals', FixedField, 'u8'),
        ('p1_b_num_signals', FixedField, 'u8'),
        ('p2_a_num_signals', FixedField, 'u8'),
        ('p2_b_num_signals', FixedField, 'u8'),
        ('p1_gbt_num_trcv', FixedField, 'u4'),
        ('p2_gbt_num_trcv', FixedField, 'u4'),
        ('tck_max_clock', FixedField, 'u8')
    ]


@fmc_multirecord(0x01)
class FmcPlusMainDefinition(FmcEntry):
    ''' ANSI/VITA 57.4-2018 FMC+ Standard, Table 5.3.1-1 '''

    _schema = [
        ('module_size', FixedField, 'u2', {'constants': {
            'single_width': 0b00,
            'double_width': 0b01
        }}),
        ('p1_p3_connector_size', FixedField, 'u2', {'constants': {
            'lpc': 0b00,
            'hpc': 0b01,
            'hspc': 0b10,
            'p1_hspc_p3_hspce': 0b11,
        }}),
        ('p2_p4_connector_size', FixedField, 'u3', {'constants': {
            'lpc': 0b000,
            'hpc': 0b001,
            'hspc': 0b010,
            'p2_hspc_p4_hspce': 0b011,
            'not_fitted': 0b111,
        }}),
        ('clock_direction', FixedField, 'u1', {'constants': {
            'm2c': 0b0,
            'c2m': 0b1,
        }}),
        ('p1_a_num_signals', FixedField, 'u8'),
        ('p2_a_num_signals', FixedField, 'u8'),
        ('_p2_b_num_sig_1_0', FixedField, 'u2'),
        ('p1_b_num_signals', FixedField, 'u6'),
        ('_p1_gbt_num_trcv_3_0', FixedField, 'u4'),
        ('_p2_b_num_sig_5_2', FixedField, 'u4'),
        ('p2_gbt_num_trcv', FixedField, 'u6'),
        ('_p1_gbt_num_trcv_5_4', FixedField, 'u2'),
        ('tck_max_clock', FixedField, 'u8')
    ]

    # Convert FMC+ bit-twiddled fields to / from plain values

    def to_dict(self):
        result = super().to_dict()
        # re-order dict items so the result looks nice
        p2nt, tck = result['p2_gbt_num_trcv'], result['tck_max_clock']
        del result['p2_gbt_num_trcv']
        del result['tck_max_clock']
        result['p2_b_num_signals'] = ord(bitstruct.pack(
            'u2u4u2',
            0,
            self['_p2_b_num_sig_5_2'],
            self['_p2_b_num_sig_1_0']
        ))
        result['p1_gbt_num_trcv'] = ord(bitstruct.pack(
            'u2u2u4',
            0,
            self['_p1_gbt_num_trcv_5_4'],
            self['_p1_gbt_num_trcv_3_0']
        ))
        result['p2_gbt_num_trcv'] = p2nt
        result['tck_max_clock'] = tck
        return result

    def update(self, val):
        _, self['_p2_b_num_sig_5_2'], self['_p2_b_num_sig_1_0'] = bitstruct.unpack(
            'u2u4u2',
            int.to_bytes(val['p2_b_num_signals'], 1, byteorder='little')
        )
        _, self['_p1_gbt_num_trcv_5_4'], self['_p1_gbt_num_trcv_3_0'] = bitstruct.unpack(
            'u2u2u4',
            int.to_bytes(val['p1_gbt_num_trcv'], 1, byteorder='little')
        )
        del val['p2_b_num_signals']
        del val['p1_gbt_num_trcv']
        super().update(val)


@fmc_multirecord(0x10)
class FmcI2cDeviceDefinition(FmcEntry):
    ''' ANSI/VITA 57.1 FMC Standard, Table 9 '''

    _schema = [
        ('_device_string', BytearrayField, None, {'hex': False})
    ]

    _addr_encoding_lookup = bidict({
        b'!': 0b0000,
        b'"': 0b0001,
        b'#': 0b0010,
        b'$': 0b0011,
        b'%': 0b0100,
        b'&': 0b0101,
        b'\'': 0b0110,
        b'(': 0b0111,
        b')': 0b1000,
        b'*': 0b1001,
        # 0b1010 is sḱipped - see ANSI/VITA 57.1 page 67
        b'+': 0b1011,
        b',': 0b1100,
        b'-': 0b1101,
        b'.': 0b1110,
        b'/': 0b1111,
    })

    def encode_addr(self, addr_num):
        if addr_num in self._addr_encoding_lookup.values():
            return self._addr_encoding_lookup.inverse[addr_num]
        return None

    def decode_addr(self, addr_char):
        key = bytes([addr_char])
        if key in self._addr_encoding_lookup:
            return self._addr_encoding_lookup[key]
        return None

    def to_dict(self):
        result = super().to_dict()
        plaintext = deser_6bit(
            self._dict['_device_string']._value).encode('utf-8')
        devices = []
        while len(plaintext):
            # Read as many addresses as possible
            addrs = []
            while len(plaintext) and self.decode_addr(plaintext[0]) is not None:
                addrs.append(self.decode_addr(plaintext[0]))
                plaintext = plaintext[1:]

            # Read device name, until the next device address shows up
            device_name = b''
            while len(plaintext) and self.decode_addr(plaintext[0]) is None:
                device_name += plaintext[0:1]
                plaintext = plaintext[1:]

            devices.append({
                'name': device_name.decode('utf-8').strip(),
                'addresses': addrs
            })

        result['devices'] = devices
        return result

    def update(self, val):
        encoded = b''
        for device in val['devices']:
            for addr in device['addresses']:
                addr_enc = self.encode_addr(addr)
                if addr_enc is None:
                    raise RuntimeError(f'Invalid address {addr}')
                else:
                    encoded += addr_enc
            encoded += device['name'].encode('utf-8')

        self._dict['_device_string']._value = ser_6bit(encoded.decode('utf-8'))
        del val['devices']
        super().update(val)



####################################################################################
#      ____  _____________  __    __  __ _           _____ ___   _                 #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)           #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \               #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\              #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B              #
#                                                                                  #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.                   #
# Modifications Copyright(C) 2025 Advanced Micro Devices, Inc. All rights reserved #
#                  SPDX-License-Identifier: BSD-3-Clause                           #
#                                                                                  #
####################################################################################

from frugy import __version__
from frugy.areas import CommonHeader, ChassisInfo, BoardInfo, ProductInfo
from frugy.multirecords import MultirecordArea
import frugy.multirecords_ipmi
import frugy.multirecords_picmg
import frugy.multirecords_fmc
import frugy.multirecords_xilinx
import yaml
from bidict import bidict
import os
from copy import deepcopy


# YAML formatting helpers


class YamlFlowstyleList(list):
    pass


def yaml_flowstyle_list_rep(dumper, data):
    return dumper.represent_sequence(u'tag:yaml.org,2002:seq', data, flow_style=True)


yaml.add_representer(YamlFlowstyleList, yaml_flowstyle_list_rep)


def import_log(msg):
    if not hasattr(import_log, "str"):
        import_log.str = ''
    import_log.str += msg + '\n'


class Fru:
    _area_table_lookup = bidict({
        'ChassisInfo': 'chassis_info_offs',
        'BoardInfo': 'board_info_offs',
        'ProductInfo': 'product_info_offs',
        'MultirecordArea': 'multirecord_offs',
    })

    def __init__(self, initdict=None):
        self.header = CommonHeader()
        self.areas = {}
        self.comment = ''
        if initdict is not None:
            self.update(initdict)

    def factory(self, cls_name, cls_args=None):
        map = {
            'ChassisInfo': ChassisInfo,
            'BoardInfo': BoardInfo,
            'ProductInfo': ProductInfo,
            'MultirecordArea': MultirecordArea
        }
        if cls_name not in map:
            raise ValueError(f"unknown FRU area: {cls_name}")
        # Some constructors modify their arguments; make sure to leave the user-supplied initdict intact
        return map[cls_name](deepcopy(cls_args))

    def update(self, src):
        self.comment = ''
        self.areas = {k: self.factory(k, v) for k, v in src.items()}

    def to_dict(self):
        return {k: v.to_dict() for k, v in self.areas.items()}

    def __repr__(self):
        return repr(self.to_dict())

    def serialize(self):
        self.header.reset()

        # Determine offsets for areas
        curr_offs = self.header.size_total()
        for area, offs in self._area_table_lookup.items():
            if area in self.areas:
                self.header[offs] = curr_offs
                curr_offs += self.areas[area].size_total()

        # Serialize everything
        result = self.header.serialize()
        for area in self._area_table_lookup.keys():
            if area in self.areas:
                result += self.areas[area].serialize()

        return result

    def deserialize(self, input):
        import_log.str = ''
        self.areas = {}
        self.header.deserialize(input)
        for k, v in self.header.to_dict().items():
            # Ignore "internal use area"
            # TODO: Support it as opaque byte array?
            if v and k != 'internal_use_offs':
                obj_name = self._area_table_lookup.inverse[k]
                obj = self.factory(obj_name)
                obj.deserialize(input[v:])
                self.areas[obj_name] = obj

    def load_yaml(self, fname):
        with open(fname, 'r') as infile:
            fru_dict = yaml.safe_load(infile)
        self.update(fru_dict)

    def dump_yaml_raw(self):
        def fmt_tree(part):
            ''' Set lists at edges of tree to YAML flow style '''
            if type(part) is list:
                if len(part) == 0:
                    return part, False
                part, edge_flags = zip(*[fmt_tree(elem) for elem in part])
                part = list(part)
                if all(edge_flags):
                    part = YamlFlowstyleList(part)
                return part, False
            elif type(part) is dict:
                part = {k: fmt_tree(part[k])[0] for k in part.keys()}
                return part, False
            elif type(part) is str:
                # don't collapse strings in YAML, for better readability
                return part, False
            else:
                # no dict list, or str ==> reached edge of tree
                return part, True

        yaml_dict = self.to_dict()
        yaml_dict, _ = fmt_tree(yaml_dict)
        return yaml.dump(yaml_dict, sort_keys=False)

    def postprocess_yaml(self, data):
        result = ''
        if hasattr(import_log, 'str'):
            if len(import_log.str) != 0:
                import_log.str = '\n' + import_log.str
            for line in [self.comment, *import_log.str.splitlines()]:
                result += f'# {line}\n'
        line_prev = ''
        for line in data.splitlines():
            if line.endswith(':') and not line.startswith(' '):
                # add LF before entries on root level
                result += '\n'
            if line.startswith('- type:') and line_prev != 'MultirecordArea:':
                # add LF between multirecord entries
                result += '\n'
            result += line + '\n'
            line_prev = line
        return result

    def dump_yaml(self):
        yaml_raw = self.dump_yaml_raw()
        return self.postprocess_yaml(yaml_raw)

    def save_yaml(self, fname):
        with open(fname, 'w') as outfile:
            outfile.write(self.dump_yaml())

    def load_bin(self, fname):
        self.comment = f'created with frugy {__version__} from "{os.path.basename(fname)}"'
        with open(fname, 'rb') as infile:
            self.deserialize(infile.read())

    def save_bin(self, fname):
        with open(fname, 'wb') as outfile:
            outfile.write(self.serialize())



###########################################################################
#      ____  _____________  __    __  __ _           _____ ___   _        #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)  #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \      #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\     #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B     #
#                                                                         #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.          #
#                    2022 Atom Computing, Inc                             #
#                  SPDX-License-Identifier: BSD-3-Clause                  #
#                                                                         #
###########################################################################

from frugy.types import FruAreaBase, FixedField, FixedStringField, GuidField, ArrayField, BytearrayField, IpV4Field, bin2hex_helper, _grouper
from frugy.multirecords import ipmi_multirecord, MultirecordEntry
from frugy.fru_registry import FruRecordType, rec_register, rec_lookup_by_id

import re
import logging


@ipmi_multirecord(0xc0)
class PicmgEntry(MultirecordEntry):
    ''' PICMG AMC.0 Specification R2.0 '''
    ''' Superclass of all PICMG OEM multirecords '''

    _picmg_identifier = 0x315a

    def format_version(self):
        return 0x00

    def _payload_prologue(self):
        return self._picmg_identifier.to_bytes(3, 'little') \
            + self._record_id.to_bytes(length=1, byteorder='little') \
            + self.format_version().to_bytes(length=1, byteorder='little')

    @classmethod
    def from_payload(cls, payload):
        picmg_id, payload = payload[:3], payload[3:]
        rec_id, rec_fmt_version, payload = payload[0], payload[1], payload[2:]
        picmg_id = int.from_bytes(picmg_id, 'little')

        if picmg_id != cls._picmg_identifier:
            raise ValueError(
                f"PICMG identifier mismatch: expected 0x{cls._picmg_identifier:06x}, received 0x{picmg_id:06x} ({picmg_id})")

        if rec_fmt_version not in [0x00, 0x01]:
            raise RuntimeError(
                f"Unexpected record format version: 0x{rec_fmt_version:02x}")

        try:
            cls_inst = rec_lookup_by_id(
                FruRecordType.picmg_multirecord, rec_id)()
        except KeyError:
            raise RuntimeError(f"Unknown PICMG entry 0x{rec_id:02x}")

        if len(payload) == 0:
            raise EOFError(
                f"Skipping creation of {cls_inst.__class__.__name__} due to empty payload")

        cls_inst._deserialize(payload)
        return cls_inst


# PICMG AMC.0 multirecords

def picmg_multirecord(rec_id):
    def register_and_set_id(cls):
        cls._record_id = rec_id
        rec_register(cls, FruRecordType.picmg_multirecord, rec_id)
        return cls
    return register_and_set_id


def picmg_secondary(cls):
    rec_register(cls, FruRecordType.picmg_secondary)
    return cls


@picmg_multirecord(0x16)
class ModuleCurrentRequirements(PicmgEntry):
    ''' PICMG AMC.0 Specification R2.0, Table 3-10 '''

    _schema = [
        ('current_draw', FixedField, 'u8', {'div': 0.1})
    ]

# Array entry classes for PointToPointConnectivity


@picmg_secondary
class AmcChannelDescriptor(FruAreaBase):
    ''' PICMG AMC.0 Specification R2.0, Table 3-17 '''

    _lanes = [f'_lane{n}_port' for n in range(4)]
    _lane_unused = 0b11111

    _schema = [
        ('_reserved', FixedField, 'u4', {'default': 0b1111}),
        ('_lane3_port', FixedField, 'u5', {'default': _lane_unused}),
        ('_lane2_port', FixedField, 'u5', {'default': _lane_unused}),
        ('_lane1_port', FixedField, 'u5', {'default': _lane_unused}),
        ('_lane0_port', FixedField, 'u5', {'default': _lane_unused}),
    ]
    _mergeBitfield = True

    def to_dict(self):
        return [self[l] for l in AmcChannelDescriptor._lanes if self[l] != AmcChannelDescriptor._lane_unused]

    def update(self, val):
        for i, l in enumerate(AmcChannelDescriptor._lanes):
            self[l] = val[i] if i < len(
                val) else AmcChannelDescriptor._lane_unused


@picmg_secondary
class AmcLinkDescriptor(FruAreaBase):
    ''' PICMG AMC.0 Specification R2.0, Table 3-19 '''

    _lane_flag_names = [f'_lane{n}_flag' for n in range(4)]

    _link_type_standard_constants = {
        'pcie': 0x02,
        'pcie_advanced': 0x03,
        'pci_advanced_1': 0x04,
        'ethernet': 0x05,
        'serial_rapidio': 0x06,
        'storage': 0x07
    }
    _link_type_oem_constants = {
        f'oem_guid_{n}': n+0xf0 for n in range(15)
    }

    _schema = [
        ('_reserved', FixedField, 'u6', {'default': 0b111111}),
        ('asymm_match', FixedField, 'u2', {'constants': {
            'match_exact': 0b00,
            'match_10b': 0b01,
            'match_01b': 0b10
        }}),
        ('grouping_id', FixedField, 'u8'),
        ('link_type_ext', FixedField, 'u4', {'default': 0}),
        ('link_type', FixedField, 'u8', {'constants': {
            **_link_type_standard_constants,
            **_link_type_oem_constants
        }}),
        ('_lane3_flag', FixedField, 'u1'),
        ('_lane2_flag', FixedField, 'u1'),
        ('_lane1_flag', FixedField, 'u1'),
        ('_lane0_flag', FixedField, 'u1'),
        ('channel_id', FixedField, 'u8'),
    ]
    _mergeBitfield = True

    def to_dict(self):
        result = super().to_dict()
        result['lane_flags'] = [self[f] for f in self._lane_flag_names]
        return result

    def update(self, val):
        for n, f in enumerate(self._lane_flag_names):
            self[f] = val['lane_flags'][n]
        del val['lane_flags']
        super().update(val)


@picmg_multirecord(0x19)
class PointToPointConnectivity(PicmgEntry):
    ''' PICMG AMC.0 Specification R2.0, Table 3-16 '''

    _schema = [
        ('_guid_count', FixedField, 'u8', {'default': 0}),
        ('guids', ArrayField, GuidField, {'num_elems_field': '_guid_count'}),
        ('record_type', FixedField, 'u1', {'constants': {
            'amc_module': 1,
            'on_carrier_device': 0
        }}),
        ('_reserved', FixedField, 'u3', {'default': 0}),
        ('connected_dev_id', FixedField, 'u4', {'default': 0}),
        ('_channel_desc_count', FixedField, 'u8', {'default': 0}),
        ('channel_descriptors', ArrayField, AmcChannelDescriptor,
         {'num_elems_field': '_channel_desc_count'}),
        ('link_descriptors', ArrayField, AmcLinkDescriptor),
    ]


# Array entry classes for ClockConfig

_clock_id_constants = {
    # Predefined Clock IDs for AMC clocks, Table 3-33
    # TODO: Are these enough, or do we also need the ATCA Backplane clocks (Table 3-34)?
    'TCLKA': 1,
    'TCLKB': 2,
    'TCLKC': 3,
    'TCLKD': 4,
    'FCLKA': 5,
}

_resource_type_constants = {
    # Clock Resource IDs, Table 3-31
    'on_carrier': 0b00,
    'amc_module': 0b01,
    'backplane': 0b10
}


@picmg_secondary
class DirectClockDescriptor(FruAreaBase):
    ''' PICMG AMC.0 Specification R2.0, Table 3-38 '''

    _schema = [
        ('_reserved', FixedField, 'u6', {'default': 0}),
        ('pll_connect', FixedField, 'u1'),
        ('asymm_match', FixedField, 'u1', {'constants': {
            'clk_source': 1,
            'clk_receiver': 0
        }}),
        ('family', FixedField, 'u8', {'constants': {
            'unspecified': 0,
            'sonet_sdh_pdh': 1,
            'pcie_reserved': 2
        }}),
        ('accuracy', FixedField, 'u8'),
        ('frequency', FixedField, 'u32'),
        ('freq_min', FixedField, 'u32'),
        ('freq_max', FixedField, 'u32'),
    ]


@picmg_secondary
class IndirectClockDescriptor(FruAreaBase):
    ''' PICMG AMC.0 Specification R2.0, Table 3-37 '''

    _schema = [
        ('_reserved', FixedField, 'u6', {'default': 0}),
        ('pll_connect', FixedField, 'u1'),
        ('asymm_match', FixedField, 'u1', {'constants': {
            'clk_src': 1,
            'clk_recv': 0
        }}),
        ('dep_clk_id', FixedField, 'u8'),
    ]


@picmg_secondary
class ClockConfigDescriptor(FruAreaBase):
    ''' PICMG AMC.0 Specification R2.0, Table 3-36 '''

    _schema = [
        ('clk_id', FixedField, 'u8', {'constants': _clock_id_constants}),
        ('_reserved', FixedField, 'u7', {'default': 0}),
        ('activation', FixedField, 'u1', {'constants': {
            'by_carrier': 0,
            'by_application': 1
        }}),
        ('_indirect_clk_desc_count', FixedField, 'u8', {'default': 0}),
        ('_direct_clk_desc_count', FixedField, 'u8', {'default': 0}),
        ('indirect_clk_desc', ArrayField, IndirectClockDescriptor,
         {'num_elems_field': '_indirect_clk_desc_count'}),
        ('direct_clk_desc', ArrayField, DirectClockDescriptor,
         {'num_elems_field': '_direct_clk_desc_count'}),
    ]


@picmg_multirecord(0x2d)
class ClockConfig(PicmgEntry):
    ''' PICMG AMC.0 Specification R2.0, Table 3-35 '''

    _schema = [
        ('resource_type', FixedField, 'u2', {
         'constants': _resource_type_constants}),
        ('_reserved', FixedField, 'u2', {'default': 0}),
        ('dev_id', FixedField, 'u4'),
        ('_conf_desc_count', FixedField, 'u8', {'default': 0}),
        ('conf_desc', ArrayField, ClockConfigDescriptor,
         {'num_elems_field': '_conf_desc_count'}),
    ]


@picmg_multirecord(0x30)
class Zone3InterfaceCompatibility(PicmgEntry):
    ''' PICMG MicroTCA.4 Enhancements for Rear I/O and Timing R1.0, Table 3-3 '''
    ''' Types 0..4 are treated as opaque hex blob. Type 5 (CLASS_ID) is decoded to plain text. '''

    _schema = [
        ('identifier_type', FixedField, 'u8', {'constants': {
            'PICMG_IRTM0_REP': 0,
            'PICMG_OTHER': 1,
            'GUID': 2,
            'OEM': 3,
            'MTCA4_REP': 4,
            'CLASS_ID': 5
        }}),
        ('identifier_body', BytearrayField, None, {'hex': True}),
    ]

    # Record Format Version - this record is special, it has 0x01 version
    def format_version(self):
        return 0x01

    # Conversion from binary data to plain-text dictionary
    def to_dict(self):
        result = super().to_dict()
        if result['identifier_type'] == 'CLASS_ID':
            # Convert hex string to raw data
            raw = bytearray.fromhex(result['identifier_body'])
            # Convert raw data to plain text
            # Skip number of elements
            raw = raw[1:]
            plain = []
            for tup in _grouper(3, raw):
                ad = ['A', 'D'][tup[0]]
                plain.append(f'{ad}{tup[1]}.{tup[2]}')
            result['identifier_body'] = plain
        return result

    # Conversion from plain-text dictionary to binary data
    def update(self, val):
        if val['identifier_type'] == 'CLASS_ID':
            cls_ver = re.compile(r'^(A|D)(\d+).(\d+)$')
            ib = val['identifier_body']
            # Convert plain text to raw data
            # Number of elements
            raw = bytearray([len(ib)])
            for s in ib:
                grp = cls_ver.match(s)
                el = grp.groups()
                if not grp or len(el) != 3:
                    logging.error(f"Can't parse class definition {s}")
                    continue
                ad = {'A': 0, 'D': 1}[el[0]]
                raw.extend(bytearray([ad, int(el[1]), int(el[2])]))
            # Convert raw data to hex string
            val['identifier_body'] = raw.hex()
        super().update(val)


@picmg_secondary
class PartitionDescriptor(FruAreaBase):
    ''' PICMG Specification MTCA.0 R1.0, Table 3-11 '''

    _schema = [
        ('offset', FixedField, 'u16', {'div': 0x10}),
        ('length', FixedField, 'u16'),
    ]


@picmg_multirecord(0x20)
class FruPartition(PicmgEntry):
    ''' PICMG Specification MTCA.0 R1.0, Table 3-10 '''

    _schema = [
        ('_desc_count', FixedField, 'u8', {'default': 0}),
        ('descriptors', ArrayField, PartitionDescriptor,
         {'num_elems_field': '_desc_count'}),
    ]


@picmg_multirecord(0x21)
class CarrierManagerIPLink(PicmgEntry):
    ''' PICMG Specification MTCA.0 R1.0, Table 3-12 '''
    _schema = [
        ('shelf_manager_ip', IpV4Field),
        ('carrier_manager_ip', IpV4Field),
        ('mch1_ip', IpV4Field),
        ('mch2_ip', IpV4Field),
        ('subnet', IpV4Field),
        ('gateway0', IpV4Field),
        ('gateway1', IpV4Field),
        ('username', FixedStringField, 17, {'default': ''}),
        ('password', FixedStringField, 21, {'default': ''}),
    ]


# PICMG Specification MTCA.0 R1.0 Table 3-14

_site_type_std = {
    'cooling_unit': 0x04,
    'advanced_mc': 0x07,
    'rtm': 0x09,
    'mtca_carrier_hub': 0x0a,
    'power_module': 0x0b,
    'unknown': 0xff
}
_site_type_oem = {
    f'oem_module_{n}': 0xc0 + n for n in range(16)
}
_site_type_constants = {
    **_site_type_std,
    **_site_type_oem
}


@picmg_secondary
class SlotEntry(FruAreaBase):
    ''' PICMG Specification MTCA.0 R1.0, Table 3-17 '''

    _schema = [
        ('site_no', FixedField, 'u8'),
        ('site_type', FixedField, 'u8', {'constants': _site_type_constants}),
        ('slot_no', FixedField, 'u8'),
        ('tier_no', FixedField, 'u8'),
        ('slot_org_y', FixedField, 'u16'),
        ('slot_org_x', FixedField, 'u16'),
    ]


@picmg_multirecord(0x22)
class MtcaCarrierInformation(PicmgEntry):
    ''' PICMG Specification MTCA.0 R1.0, Table 3-16 '''

    _schema = [
        ('number', FixedField, 'u8', {'default': 0xff}),
        ('orientation', FixedField, 'u1', {'constants': {
            'l2r': 0,
            'b2t': 1
        }}),
        ('_slot_entry_count', FixedField, 'u7', {'default': 0}),
        ('slot_entries', ArrayField, SlotEntry, {
         'num_elems_field': '_slot_entry_count'}),
    ]


@picmg_secondary
class PowerPolicyDescriptor(FruAreaBase):
    ''' PICMG Specification MTCA.0 R1.0, Table 3-24 '''

    _schema = [
        ('site_no', FixedField, 'u8'),
        ('max_current_override', FixedField, 'u16', {'div': 0.1}),
        ('pm_role', FixedField, 'u8', {'constants': {
            'primary': 0x00,
            'redundant': 0x01,
            'unspecified': 0xff
        }}),
        ('_channel_count', FixedField, 'u8'),
        ('_channels', BytearrayField, None, {
         'num_elems_field': '_channel_count'}),
    ]

    def to_dict(self):
        ''' Convert _channels from bytearray to list of ints '''
        result = super().to_dict()
        result['channels'] = list(self._dict['_channels']._value)
        return result

    def update(self, val):
        ''' Convert channels from list of ints to bytearray '''
        self._dict['_channels']._value = bytearray(val['channels'])
        del val['channels']
        super().update(val)


@picmg_multirecord(0x25)
class PowerPolicyRecord(PicmgEntry):
    ''' PICMG Specification MTCA.0 R1.0, Table 3-23 '''

    _schema = [
        ('_num_descriptors', FixedField, 'u8', {'default': 0}),
        ('descriptors', ArrayField, PowerPolicyDescriptor,
         {'num_elems_field': '_num_descriptors'}),
    ]


@picmg_secondary
class MtcaCarrierActivCurrDescriptor(FruAreaBase):
    ''' PICMG Specification MTCA.0 R1.0, Table 3-26 '''

    _mgr_constants = {
        'reserved': 0b11,
        'shelf_mgr': 0b10,
        'carrier_mgr': 0b01,
        'system_mgr': 0b00
    }
    _schema = [
        ('site_type', FixedField, 'u8', {'constants': _site_type_constants}),
        ('site_no', FixedField, 'u8'),
        ('pwr_ch', FixedField, 'u8'),
        ('max_current', FixedField, 'u8', {'div': 0.1}),
        ('activation_ctrl', FixedField, 'u2', {'constants': _mgr_constants}),
        ('pwr_delay', FixedField, 'u6', {'div': 0.1}),
        ('deactivation_ctrl', FixedField, 'u2', {'constants': _mgr_constants}),
        ('_reserved', FixedField, 'u6', {'default': 0})
    ]


@picmg_multirecord(0x26)
class MtcaCarrierActivationPm(PicmgEntry):
    ''' PICMG Specification MTCA.0 R1.0, Table 3-25 '''

    _schema = [
        ('readiness_allowance', FixedField, 'u8'),
        ('_num_descriptors', FixedField, 'u8', {'default': 0}),
        ('descriptors', ArrayField, MtcaCarrierActivCurrDescriptor,
         {'num_elems_field': '_num_descriptors'}),
    ]


@picmg_secondary
class P2pPortDescriptor(FruAreaBase):
    ''' PICMG AMC.0 Specification R2.0, Table 3-15 '''

    _mergeBitfield = True

    _schema = [
        ('_reserved0', FixedField, 'u6', {'default': 0}),
        ('local_port', FixedField, 'u5'),
        ('remote_port', FixedField, 'u5'),
        ('resource_type', FixedField, 'u1', {'constants': {
            'amc': 1,
            'carrier': 0
        }}),
        ('_reserved1', FixedField, 'u3', {'default': 0}),
        ('site_no', FixedField, 'u4'),
    ]


@picmg_secondary
class P2pAmcResourceDescriptor(FruAreaBase):
    ''' PICMG AMC.0 Specification R2.0, Table 3-14 '''

    _schema = [
        ('resource_type', FixedField, 'u1', {'constants': {
            'amc': 1,
            'carrier': 0
        }}),
        ('_reserved', FixedField, 'u3', {'default': 0}),
        ('site_no', FixedField, 'u4'),
        ('_port_count', FixedField, 'u8'),
        ('port_descriptors', ArrayField, P2pPortDescriptor,
         {'num_elems_field': '_port_count'}),
    ]


@picmg_multirecord(0x18)
class CarrierP2pConnectivity(PicmgEntry):
    ''' PICMG AMC.0 Specification R2.0, Table 3-13 '''

    _schema = [
        ('resource_descriptors', ArrayField, P2pAmcResourceDescriptor),
    ]


@picmg_secondary
class P2pClockConnectionDescriptor(FruAreaBase):
    ''' PICMG AMC.0 Specification R2.0, Table 3-32 '''

    _schema = [
        ('local_clock_id', FixedField, 'u8',
         {'constants': _clock_id_constants}),
        ('remote_clock_id', FixedField, 'u8',
         {'constants': _clock_id_constants}),
        ('resource_type', FixedField, 'u2', {
         'constants': _resource_type_constants}),
        ('_reserved', FixedField, 'u2', {'default': 0}),
        ('dev_id', FixedField, 'u4'),
    ]


@picmg_secondary
class ClockP2pResourceDescriptor(FruAreaBase):
    ''' PICMG AMC.0 Specification R2.0, Table 3-30 '''

    _schema = [
        ('resource_type', FixedField, 'u2', {
         'constants': _resource_type_constants}),
        ('_reserved', FixedField, 'u2', {'default': 0}),
        ('dev_id', FixedField, 'u4'),
        ('_p2p_clk_conn_count', FixedField, 'u8', {'default': 0}),
        ('p2p_clk_conn_descriptors', ArrayField, P2pClockConnectionDescriptor,
         {'num_elems_field': '_p2p_clk_conn_count'})
    ]


@picmg_multirecord(0x2c)
class CarrierClkP2pConnectivity(PicmgEntry):
    ''' PICMG AMC.0 Specification R2.0, Table 3-29 '''

    _schema = [
        ('_clk_p2p_resource_desc_count', FixedField, 'u8', {'default': 0}),
        ('clk_p2p_resource_descriptors', ArrayField, ClockP2pResourceDescriptor, {
         'num_elems_field': '_clk_p2p_resource_desc_count'}),
    ]


@picmg_secondary
class BusedDeviceDescriptor(FruAreaBase):
    ''' PICMG MicroTCA.4 Enhancements for Rear I/O and Timing R1.0, Table 3-14 '''

    _resource_id_constants = {
        'on_carrier': 0b0,
        'AMC': 0b1,
    }

    _schema = [
        ('resource_id', FixedField, 'u1', {'constants': _resource_id_constants}),
        ('_rsvd', FixedField, 'u3', {'default': 0}),
        ('amc_site', FixedField, 'u4'),
        ('port', FixedField, 'u8'),
    ]


@picmg_secondary
class BusedConnectionDescriptor(FruAreaBase):
    ''' PICMG MicroTCA.4 Enhancements for Rear I/O and Timing R1.0, Table 3-13 '''
    _schema = [
        ('_bused_device_count', FixedField, 'u8'),
        ('bused_device_descriptor', ArrayField, BusedDeviceDescriptor,
            {'num_elems_field': '_bused_device_count'}),
    ]


@picmg_multirecord(0x31)
class CarrierBusedConnectivity(PicmgEntry):
    ''' PICMG MicroTCA.4 Enhancements for Rear I/O and Timing R1.0, Table 3-12 '''
    _schema = [
        ('_count', FixedField, 'u8'),
        ('bused_connection_descriptors', ArrayField, BusedConnectionDescriptor,
            {'num_elems_field': '_count'})
    ]


@picmg_multirecord(0x32)
class Zone3InterfaceDocumentation(PicmgEntry):
    ''' PICMG MicroTCA.4 Enhancements for Rear I/O and Timing R1.0, Table 3-15 '''

    _schema = [
        ('url', BytearrayField, None, {'hex': False}),
    ]



####################################################################################
#      ____  _____________  __    __  __ _           _____ ___   _                 #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)           #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \               #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\              #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B              #
#                                                                                  #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.                   #
# Modifications Copyright(C) 2025 Advanced Micro Devices, Inc. All rights reserved #
#                  SPDX-License-Identifier: BSD-3-Clause                           #
#                                                                                  #
####################################################################################

from enum import Enum, auto
from collections import defaultdict
from itertools import chain


class FruRecordType(Enum):
    ipmi_area = auto()
    ipmi_multirecord = auto()
    picmg_multirecord = auto()
    picmg_secondary = auto()
    fmc_multirecord = auto()
    fmc_secondary = auto()
    xilinx_multirecord = auto()


_registry = defaultdict(list)

_lookup_by_id = defaultdict(dict)
_lookup_by_name = {}


def rec_register(cls, rec_type: FruRecordType, rec_id=None):
    ''' Register FRU record type in central registry and lookup table '''
    _registry[rec_type].append(cls)
    _lookup_by_name[cls.__name__] = cls
    if rec_id is not None:
        _lookup_by_id[rec_type][rec_id] = cls


def rec_enumerate(rec_type_filter=None):
    ''' Enumerate registered FRU records, filter if rec_type_filter given '''
    if rec_type_filter is None:
        rec_type_filter = list(FruRecordType)
    elif type(rec_type_filter) is not list:
        rec_type_filter = [rec_type_filter]

    result = [_registry[rec_type] for rec_type in rec_type_filter]

    return list(chain.from_iterable(result))


def rec_lookup_by_id(rec_type: FruRecordType, rec_id: int):
    ''' Lookup record class by record type and ID '''
    return _lookup_by_id[rec_type][rec_id]


def rec_lookup_by_name(rec_name: str):
    ''' Lookup record class by name '''
    return _lookup_by_name[rec_name]


def rec_info(rec):
    ''' return: record name, record information(docstring) '''
    return rec.__name__, str(rec.__doc__).strip()


def schema_entry_info(entry):
    ''' return schema entry name, type, options (e.g. constants) '''
    e_name = entry[0]
    e_inst = entry[1]._shortname
    if e_inst == 'array':
        e_args = f'({entry[2].__name__})'
    else:
        e_args = f'({entry[2]})' if len(
            entry) > 2 and entry[2] is not None else ''
    e_opt = None
    if len(entry) > 3:
        e_opt = entry[3]

    e_type = f'{e_inst} {e_args}'

    return e_name, e_type, e_opt



####################################################################################
#      ____  _____________  __    __  __ _           _____ ___   _                 #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)           #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \               #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\              #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B              #
#                                                                                  #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.                   #
# Modifications Copyright(C) 2025 Advanced Micro Devices, Inc. All rights reserved #
#                  SPDX-License-Identifier: BSD-3-Clause                           #
#                                                                                  #
####################################################################################

import bitstruct
from collections import OrderedDict
from enum import Enum
from itertools import zip_longest
import uuid
from bidict import bidict
from ipaddress import IPv4Address
from macaddress import MAC
import logging

_format_version_default = 1
_en_decode='ISO-8859-1'

def _sizeAlign(size: int, alignment: int) -> int:
    ''' return number of padding bytes & total length after padding '''
    numPadBytes = -size % alignment
    return numPadBytes, size + numPadBytes


def _grouper(n, iterable, padvalue=None):
    "grouper(3, 'abcdefg', 'x') -> ('a','b','c'), ('d','e','f'), ('g','x','x')"
    return zip_longest(*[iter(iterable)]*n, fillvalue=padvalue)


def ser_6bit(val: str) -> bytearray:
    result = b''
    for chunk in _grouper(4, val.upper(), padvalue=' '):
        chunk = list(map(lambda x: ord(x) - 0x20, chunk))
        chunk.reverse()
        tmp = bitstruct.pack('u6'*4, *chunk)
        result += tmp[::-1]
    return result


def deser_6bit(val: bytearray) -> str:
    result = b''
    for chunk in _grouper(3, val, padvalue=' '):
        tmp = bitstruct.unpack('u6'*4, bytearray(reversed(chunk)))
        for x in tmp[::-1]:
            result += bytes((x + 0x20,))
    return result.decode(_en_decode)


def bin2hex_helper(val: bytearray):
    return ' '.join('%02x' % x for x in val)


class FixedField():
    ''' Fixed length field for numbers & bitfields '''
    _shortname = 'int'

    def __init__(self, format: str, parent=None, default=None, div=None, constants=None):
        self._format = format
        self._default = default
        self._value = default
        self._div = div
        self._constants_lookup = bidict(
            constants) if constants is not None else None

    def bit_fmt(self) -> str:
        return self._format

    def bit_size(self) -> int:
        return bitstruct.calcsize(self._format)

    def to_serialized(self):
        tmp = self._value
        if self._div is not None:
            tmp = int(tmp / self._div)
        return tmp

    def from_serialized(self, value):
        if self._div is not None:
            if self._div < 1:
                # Stupid hack that avoids rounding issues a la 7.6000000000000005
                value /= 1 / self._div
            else:
                value *= self._div
        self._value = value

    def to_dict(self):
        if self._constants_lookup is not None and self._value in self._constants_lookup.inverse:
            return self._constants_lookup.inverse[self._value]
        else:
            return self._value

    def update(self, value):
        if self._constants_lookup is not None and value in self._constants_lookup:
            self._value = self._constants_lookup[value]
        else:
            self._value = value

    def val_not_default(self):
        return self.to_dict() != self._default


class StringFmt(Enum):
    BIN = 0b00
    BCD_PLUS = 0b01
    ASCII_6BIT = 0b10
    ASCII_8BIT = 0b11


class StringField():
    ''' Variable length field for strings'''
    _shortname = 'str'

    def __init__(self, default='', format: StringFmt = StringFmt.ASCII_8BIT, parent=None):
        if default is None:
            print(f'ERROR: {parent.__class__.__name__}')
        self._format = format
        self._default = default
        self._value = default

    bcdplus_lookup = bidict({
        '0': 0, '1': 1, '2': 2, '3': 3, '4': 4, '5': 5, '6': 6, '7': 7,
        '8': 8, '9': 9, ' ': 10, '-': 11, '.': 12
    })

    def bit_size(self) -> int:
        def size_bin(val: str) -> int:
            return int(len(val) / 2)

        def size_plain(val: str) -> int:
            return len(val)

        def size_6bit(val: str) -> int:
            _, n = _sizeAlign(len(val), 4)
            return (n // 4) * 3

        def size_bcd_plus(val: str) -> int:
            _, n = _sizeAlign(len(val), 2)
            return n // 2

        size_fn = {
            StringFmt.BIN: size_bin,
            StringFmt.BCD_PLUS: size_bcd_plus,
            StringFmt.ASCII_6BIT: size_6bit,
            StringFmt.ASCII_8BIT: size_plain
        }[self._format]
        return (size_fn(self._value) + 1) * 8

    def size_total(self) -> int:
        return self.bit_size() // 8

    def serialize(self) -> bytearray:
        def ser_bin(val: str) -> bytearray:
            try:
                val = bytearray(bytes.fromhex(val))
            except:
                val = b""
            return val

        def ser_plain(val: str) -> bytearray:
            return val.encode(_en_decode)

        def ser_bcd_plus(val: str) -> bytearray:
            result = b''
            for chunk in _grouper(2, val, padvalue=' '):
                chunk = map(lambda x: self.bcdplus_lookup[x], chunk)
                result += bitstruct.pack('u4'*2, *chunk)
            return result

        def ser_type_length(val: str) -> int:
            return bitstruct.pack('u2u6', self._format.value, len(val))

        ser_fn = {
            StringFmt.BIN: ser_bin,
            StringFmt.BCD_PLUS: ser_bcd_plus,
            StringFmt.ASCII_6BIT: ser_6bit,
            StringFmt.ASCII_8BIT: ser_plain
        }[self._format]
        result = ser_fn(self._value)
        return ser_type_length(result) + result

    def deserialize(self, input: bytearray) -> bytearray:
        def deser_bin(val: bytearray) -> str:
            strval = str(val)
            return val.hex()

        def deser_plain(val: bytearray) -> str:
            return val.decode(_en_decode)

        def deser_bcd_plus(val: bytearray) -> str:
            result = ''
            for v in val:
                tmp = bitstruct.unpack('u4'*2, bytes((v,)))
                for x in tmp:
                    result += self.bcdplus_lookup.inverse[x]
            return result

        fmt_int, payload_len = bitstruct.unpack('u2u6', input[0:1])
        self._format = StringFmt(fmt_int)
        logging.debug(
            f'{self.__class__.__name__}: string format: {self._format.name}, length: {payload_len}')
        remainder = input[1:]
        payload, remainder = remainder[:payload_len], remainder[payload_len:]

        deser_fn = {
            StringFmt.BIN: deser_bin,
            StringFmt.BCD_PLUS: deser_bcd_plus,
            StringFmt.ASCII_6BIT: deser_6bit,
            StringFmt.ASCII_8BIT: deser_plain
        }[self._format]
        self._value = deser_fn(payload)

        logging.debug(
            f'{self.__class__.__name__}: val: {self._value}, remainder: {bin2hex_helper(remainder[:10])}...')
        return remainder

    def to_dict(self):
        return self._value

    def format(self):
        return self._format

    def update(self, value):
        self._value = value

    def val_not_default(self):
        return self.to_dict() != self._default


class BytearrayField():
    ''' Variable length field containing a transparent bytearray, to handle stupid ambiguous multirecord payloads '''
    ''' (e.g. "Zone 3 Interface Compatibility record") '''
    _shortname = 'bytes'

    def __init__(self, default='', num_elems_field=None, hex=True, parent=None):
        self._parent = parent
        self._num_elems_field = num_elems_field
        self._default = default
        self._value = default
        self._hex = hex

    def bit_size(self) -> int:
        return len(self._value) * 8

    def pre_serialize(self):
        ''' Set number of elements in parent record, _before_ serializing it '''
        if self._num_elems_field:
            self._parent._set(self._num_elems_field, len(self._value))

    def serialize(self) -> bytearray:
        return self._value

    def deserialize(self, input: bytearray) -> bytearray:
        if self._num_elems_field:
            num_elems = self._parent._get(self._num_elems_field)
            self._value = input[:num_elems]
            return input[num_elems:]
        else:
            self._value = input
            return b''

    def to_dict(self):
        return bin2hex_helper(self._value) if self._hex else self._value.decode(_en_decode)

    def update(self, value):
        self._value = bytearray.fromhex(
            value) if self._hex else value.encode(_en_decode)

    def val_not_default(self):
        return self.to_dict() != self._default


class FixedStringField():
    ''' Null-terminated string with fixed size buffer '''
    _shortname = 'str'

    _null_term = b'\x00'

    def __init__(self, bufsize, default='', parent=None):
        self._bufsize = bufsize
        self._default = default
        self._value = default

    def bit_size(self) -> int:
        return self._bufsize * 8

    def serialize(self) -> bytearray:
        return self._value[:self._bufsize-1].encode(_en_decode) + self._null_term

    def deserialize(self, input: bytearray) -> bytearray:
        tmp, remainder = input[:self._bufsize], input[self._bufsize:]
        if self._null_term in tmp:
            pos = tmp.index(self._null_term)
            tmp = tmp[:pos]
        self._value = tmp.decode(_en_decode)
        return remainder

    def to_dict(self):
        return self._value

    def update(self, value):
        self._value = value

    def val_not_default(self):
        return self.to_dict() != self._default


class CustomStringArray:
    ''' Platform Management FRU Information Storage Definition, Table 10-1, 11-1, 12-1 '''
    ''' This is a daisy-chain of StringField objects, delimited with 0xc1 '''
    _shortname = 'strarray'

    _delimiter = b'\xc1'

    def __init__(self, initdict=None, parent=None):
        self.strings = []
        if initdict is not None:
            self.update(initdict)

    def update(self, initdict):
        try:
            for v, k in initdict.items():
                if v == "ASCII_8BIT":
                    for s in k:
                        self.strings += [StringField(s, format = StringFmt.ASCII_8BIT)]
                elif v == "BIN":
                    for s in k:
                        self.strings += [StringField(s, format = StringFmt.BIN)]
                elif v == "BCD_PLUS":
                    for s in k:
                        self.strings += [StringField(s, format = StringFmt.BCD_PLUS)]
                elif v == "ASCII_6BIT":
                    for s in k:
                        self.strings += [StringField(s, format = StringFmt.ASCII_6BIT)]
        except:
            self.strings = [StringField(v, format = StringFmt.BIN) for v in initdict]

    def __repr__(self):
        return self.to_dict().__repr__()

    def to_dict(self):
        ret = {}
        for v in self.strings:
            if v.format() == StringFmt.ASCII_8BIT:
                if "ASCII_8BIT" in ret:
                    ret["ASCII_8BIT"] += [v.to_dict()]
                else:
                    ret["ASCII_8BIT"] = [v.to_dict()]
            elif v.format() == StringFmt.ASCII_6BIT:
                if "ASCII_6BIT" in ret:
                    ret["ASCII_6BIT"] += [v.to_dict()]
                else:
                    ret["ASCII_6BIT"] = [v.to_dict()]
            elif v.format() == StringFmt.BCD_PLUS:
                if "BCD_PLUS" in ret:
                    ret["BCD_PLUS"] += [v.to_dict()]
                else:
                    ret["BCD_PLUS"] = [v.to_dict()]
            elif v.format() == StringFmt.BIN:
                if "BIN" in ret:
                    ret["BIN"] += [v.to_dict()]
                else:
                    ret["BIN"] = [v.to_dict()]

        return ret

    def serialize(self):
        result = b''
        for v in self.strings:
            result += v.serialize()
        return result + self._delimiter

    def deserialize(self, input):
        self.strings = []
        remainder = input
        while len(remainder):
            if remainder[0:1] == self._delimiter:
                remainder = remainder[1:]
                break
            tmp = StringField(format = StringFmt.BIN)
            remainder = tmp.deserialize(remainder)
            self.strings.append(tmp)

        logging.debug(
            f'{self.__class__.__name__}: parsed {len(self.strings)} strings')
        return remainder

    def size_total(self):
        # add one for the delimiter
        return sum([v.size_total() for v in self.strings]) + 1

    def bit_size(self) -> int:
        return self.size_total() * 8

    def val_not_default(self):
        return len(self.strings) != 0


class IpV4Field():
    ''' Field containing a IPv4 address '''
    _shortname = 'ipv4'

    _num_bytes = 4

    def __init__(self, default='0.0.0.0', parent=None):
        self._default = default
        self._value = IPv4Address(default)

    def bit_size(self) -> int:
        return self._num_bytes * 8

    def serialize(self) -> bytearray:
        return int(self._value).to_bytes(self._num_bytes, 'big')

    def deserialize(self, input: bytearray) -> bytearray:
        tmp, remainder = input[:self._num_bytes], input[self._num_bytes:]
        tmp = int.from_bytes(tmp, 'big')
        self._value = IPv4Address(tmp)
        return remainder

    def to_dict(self):
        return str(self._value)

    def update(self, value):
        self._value = IPv4Address(value)

    def val_not_default(self):
        return self.to_dict() != self._default


class MacField():
    ''' Field containing a IPv4 address '''
    _shortname = 'mac'

    _num_bytes = 6

    def __init__(self, default='00:00:00:00:00:00', parent=None):
        self._default = default
        if (default == None):
            self._value = None
        else:
            self._value = MAC(default)

    def bit_size(self) -> int:
        return self._num_bytes * 8

    def serialize(self) -> bytearray:
        if (self._value == None):
            return b''
        return int(self._value).to_bytes(self._num_bytes, 'big')

    def deserialize(self, input: bytearray) -> bytearray:
        tmp, remainder = input[:self._num_bytes], input[self._num_bytes:]
        tmp = int.from_bytes(tmp, 'big')
        if (tmp == 0):
            self._value = ''
        else:
            self._value = MAC(tmp)
        return remainder

    def to_dict(self):
        return str(self._value)

    def update(self, value):
        self._value = MAC(value)

    def val_not_default(self):
        return self.to_dict() != self._default


class GuidField():
    ''' Field containing a 128-bit GUID '''
    _shortname = 'guid'

    _uuid_len = 16

    def __init__(self, value=None, parent=None):
        self._value = uuid.UUID(value) if value is not None else uuid.uuid4()

    def bit_size(self) -> int:
        return GuidField._uuid_len * 8

    def size_total(self) -> int:
        return GuidField._uuid_len

    def serialize(self) -> bytearray:
        return self._value.bytes_le

    def deserialize(self, input: bytearray) -> bytearray:
        payload, remainder = input[:GuidField._uuid_len], input[GuidField._uuid_len:]
        self._value = uuid.UUID(bytes_le=payload)
        return remainder

    def to_dict(self):
        return str(self._value)

    def update(self, value):
        self._value = uuid.UUID(value)


class ArrayField():
    ''' Field containing an array of instances of another record '''
    _shortname = 'array'

    def __init__(self, cls, parent=None, initdict=None, num_elems_field=None):
        self._parent = parent
        self._cls = cls
        self._records = []
        self._num_elems_field = num_elems_field
        if initdict is not None:
            self.update(initdict)

    def update(self, initdict):
        self._records = []
        for v in initdict:
            self._records.append(self._cls(v))

    def __repr__(self):
        return self.to_dict().__repr__()

    def to_dict(self):
        return [v.to_dict() for v in self._records]

    def pre_serialize(self):
        ''' Set number of elements in parent record, _before_ serializing it '''
        if self._num_elems_field:
            self._parent._set(self._num_elems_field, self.num_elems())

    def serialize(self):
        result = b''
        for f in self._records:
            result += f.serialize()
        return result

    def deserialize(self, input):
        self._records = []
        remainder = input
        num_elems = None
        if self._num_elems_field:
            num_elems = self._parent._get(self._num_elems_field)

        while len(remainder) and num_elems != 0:
            record = self._cls()
            remainder = record.deserialize(remainder)
            self._records.append(record)
            if num_elems is not None:
                num_elems -= 1

        return remainder

    def bit_size(self):
        return self.size_total() * 8

    def size_total(self):
        return sum([v.size_total() for v in self._records])

    def num_elems(self):
        return len(self._records)

    def val_not_default(self):
        return self.num_elems() != 0


class FruAreaBase:
    ''' Common base class for FRU areas '''

    _mergeBitfield = False

    def __init__(self, initdict=None):
        self._dict = OrderedDict()
        for v in self._schema:
            kwargs = v[3] if len(v) > 3 else {}
            obj = v[1]
            if len(v) > 2:
                obj = obj(v[2], parent=self, **kwargs)
            else:
                obj = obj(parent=self, **kwargs)
            self._dict[v[0]] = obj

        if initdict is not None:
            self.update(initdict)

    # dict interface

    def __getitem__(self, key):
        # check for special accessor
        fname = f'_get_{key}'
        if hasattr(self, fname):
            return getattr(self, fname)()
        else:
            # use generic accessor
            return self._get(key)

    def __setitem__(self, key, value):
        # check for special accessor
        fname = f'_set_{key}'
        if hasattr(self, fname):
            getattr(self, fname)(value)
        else:
            # use generic accessor
            self._set(key, value)

    def __contains__(self, key):
        return hasattr(self, f'_set_{key}') or key in self._dict

    def __repr__(self):
        return repr(self.to_dict())

    def update(self, src):
        for k, v in src.items():
            self[k] = v

    def to_dict(self):
        # Fields starting with _ are ignored by convention (reserved values).
        return {
            k: self[k] for k in self._dict.keys()
            if not k.startswith('_') and self._dict[k].val_not_default()
        }

    # accessors

    def _get(self, key):
        return self._dict[key].to_dict()

    def _set(self, key, value):
        self._dict[key].update(value)

    # Fixed point integer helpers

    def _set_div(self, key, value, div):
        self._set(key, int(value / div))

    def _get_div(self, key, div):
        return float(self._get(key)) * div

    # (de)serializing

    def size_payload(self) -> int:
        return sum([v.bit_size() for v in self._dict.values()]) // 8

    def size_total(self) -> int:
        return self.size_payload()

    def _serialize(self) -> bytearray:
        for v in self._dict.values():
            if hasattr(v, 'pre_serialize'):
                v.pre_serialize()

        result = b''
        bit_fmt = ''
        bit_values = []

        def serialize_bitfield():
            nonlocal bit_fmt, bit_values
            if bit_fmt == '':
                return b''
            if not self._mergeBitfield:
                bits = bitstruct.pack(bit_fmt + '<', *bit_values)
            else:
                bits = bitstruct.pack(bit_fmt, *bit_values)[::-1]
            bit_fmt = ''
            bit_values = []
            return bits

        for v in self._dict.values():
            if hasattr(v, 'bit_fmt'):
                bit_fmt += v.bit_fmt()
                bit_values.append(v.to_serialized())
            else:
                # before any other type is serialized, serialize bitfield first
                result += serialize_bitfield()
                result += v.serialize()
        # finish serializing bitfield, if anything is left
        result += serialize_bitfield()
        return result

    def serialize(self) -> bytearray:
        return self._serialize()

    def _deserialize(self, input: bytearray):
        remainder = input
        bit_fmt = ''
        bit_fields = []

        def deserialize_bitfield():
            nonlocal bit_fmt, bit_fields, remainder
            if bit_fmt == '':
                return
            bit_length = bitstruct.calcsize(bit_fmt) // 8
            bit_data, remainder = remainder[:bit_length], remainder[bit_length:]
            logging.debug(
                f'{self.__class__.__name__}: parse {bit_fmt} from {bin2hex_helper(bit_data)}')
            if not self._mergeBitfield:
                values = bitstruct.unpack(bit_fmt + '<', bit_data)
            else:
                values = bitstruct.unpack(bit_fmt, bit_data[::-1])
            for f, v in zip(bit_fields, values):
                f.from_serialized(v)
            bit_fmt = ''
            bit_fields = []

        for k, v in self._dict.items():
            logging.debug(f'{self.__class__.__name__}: parsing {k}')
            if hasattr(v, 'bit_fmt'):
                bit_fmt += v.bit_fmt()
                bit_fields.append(v)
            else:
                # before any other type is deserialized, deserialize bitfield first
                deserialize_bitfield()
                logging.debug(
                    f'{self.__class__.__name__}: parse {v} {bin2hex_helper(remainder[:10])}...')
                remainder = v.deserialize(remainder)
        # finish deserializing bitfield, if anything is left
        deserialize_bitfield()
        return remainder

    def deserialize(self, input: bytearray):
        return self._deserialize(input)


class FruAreaChecksummed(FruAreaBase):
    ''' FRU area featuring a checksum in the epilogue '''

    ignore_checksum_errors = False

    def _prologue(self) -> bytearray:
        return b''

    def _epilogue(self, payload: bytearray) -> bytearray:
        ''' return data to append (checksum and padding to 64 bit alignment) '''
        numPadBytes, _ = _sizeAlign(len(payload) + 1, 8)
        result = b'\x00' * numPadBytes
        cksum = (-sum(payload)) & 0xff
        result += cksum.to_bytes(length=1, byteorder='little')
        return result

    def size_total(self) -> int:
        # add one byte for checksum
        _, n = _sizeAlign(self.size_payload() + 1, 8)
        return n

    def serialize(self) -> bytearray:
        payload = self._prologue()
        payload += self._serialize()
        return payload + self._epilogue(payload)

    def _verify_epilogue(self, input: bytearray, offs: int) -> bytearray:
        payload = input[:offs]
        ep = self._epilogue(payload)
        vfy, remainder = input[offs:offs+len(ep)], input[offs+len(ep):]
        if ep != vfy and not self.ignore_checksum_errors:
            raise RuntimeError(
                f'padding or checksum verify error in {self.__class__.__name__}: expected {bin2hex_helper(ep)}, received {bin2hex_helper(vfy)}')
        return remainder

    def deserialize(self, input: bytearray):
        remainder = self._deserialize(input)
        return self._verify_epilogue(input, len(input) - len(remainder))


class FruAreaVersioned(FruAreaChecksummed):
    ''' FRU area featuring a version field '''

    def __init__(self, initdict=None):
        self._format_version = _format_version_default
        super().__init__(initdict=initdict)

    def _set_format_version(self, val):
        self._format_version = val

    def _get_format_version(self):
        return self._format_version

    def _prologue(self) -> bytearray:
        ''' return data to prepend (format version) '''
        return super()._prologue() + self._format_version.to_bytes(1, 'little')

    def size_payload(self) -> int:
        # add one byte for format version
        return super().size_payload() + 1

    def deserialize(self, input: bytearray):
        self._format_version = input[0]
        remainder = self._deserialize(input[1:])
        return self._verify_epilogue(input, len(input) - len(remainder))


class FruAreaSized(FruAreaVersioned):
    ''' FRU area featuring version and length fields '''

    def _set_area_length(self, val):
        if (val % 8) != 0:
            raise ValueError(f'area length {val} not 64-bit aligned')
        self._area_length = val // 8

    def _get_area_length(self):
        return self._area_length * 8

    def size_payload(self) -> int:
        # add one byte for size field
        return super().size_payload() + 1

    def _prologue(self) -> bytearray:
        ''' return data to prepend (size field) '''
        self._set_area_length(self.size_total())
        return super()._prologue() + self._area_length.to_bytes(1, 'little')

    def deserialize(self, input: bytearray):
        self._format_version = input[0]
        self._area_length = input[1]
        area_len = self._get_area_length()
        logging.debug(
            f'{self.__class__.__name__}: parse {bin2hex_helper(input[:10])}...')
        self._deserialize(input[2:area_len])
        # The remainder may have arbitrary padding, so just check the last byte, and return only stuff that's behind our area
        cksum = (-sum(input[:area_len])) & 0xff
        if cksum != 0 and not self.ignore_checksum_errors:
            raise RuntimeError(
                f'{self.__class__.__name__}: checksum doesn\'t add up to zero')
        return input[area_len:]



###########################################################################
#      ____  _____________  __    __  __ _           _____ ___   _        #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)  #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \      #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\     #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B     #
#                                                                         #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.          #
#                  SPDX-License-Identifier: BSD-3-Clause                  #
#                                                                         #
###########################################################################

from frugy.types import FixedField, BytearrayField
from frugy.multirecords import ipmi_multirecord, MultirecordEntry


# IPMI standard multirecords

fmc_voltages_per_port = [
    'VADJ',
    '3P3V',
    '12P0V',
    'VIO_B_M2C',
    'VREF_A_M2C',
    'VREF_B_M2C'
]
fmc_voltages_total = [f'{p}_{v}' for p in ['P1', 'P2']
                      for v in fmc_voltages_per_port]
fmc_output_constants = {name: idx for idx,
                        name in enumerate(fmc_voltages_total)}


@ipmi_multirecord(0x01)
class DCOutput(MultirecordEntry):
    ''' Platform Management FRU Information Storage Definition, Table 18-2 '''

    # TODO: add unit metadata to fields
    _schema = [
        ('standby_enable', FixedField, 'u1'),
        ('_reserved', FixedField, 'u3', {'default': 0}),
        ('output_number', FixedField, 'u4', {
         'constants': fmc_output_constants}),
        ('nominal_voltage', FixedField, 'u16', {'div': 10}),     # 10mV
        ('max_neg_voltage', FixedField, 'u16', {'div': 10}),     # 10mV
        ('max_pos_voltage', FixedField, 'u16', {'div': 10}),     # 10mV
        ('max_noise_pk2pk', FixedField, 'u16'),     # mV
        ('min_current_draw', FixedField, 'u16'),    # mA
        ('max_current_draw', FixedField, 'u16'),    # mA
    ]


@ipmi_multirecord(0x02)
class DCLoad(MultirecordEntry):
    ''' Platform Management FRU Information Storage Definition, Table 18-4 '''

    _schema = [
        ('_reserved', FixedField, 'u4', {'default': 0}),
        ('output_number', FixedField, 'u4', {
         'constants': fmc_output_constants}),
        ('nominal_voltage', FixedField, 'u16', {'div': 10}),     # 10mV
        ('min_voltage', FixedField, 'u16', {'div': 10}),         # 10mV
        ('max_voltage', FixedField, 'u16', {'div': 10}),         # 10mV
        ('max_noise_pk2pk', FixedField, 'u16'),     # mV
        ('min_current_load', FixedField, 'u16'),    # mA
        ('max_current_load', FixedField, 'u16'),    # mA
    ]


@ipmi_multirecord(0x03)
class MgmtAccessRecord(MultirecordEntry):
    ''' Platform Management FRU Information Storage Definition, Table 18-6 '''

    _schema = [
        ('id', FixedField, 'u8', {'constants': {
            'sys_mgmt_url': 1,
            'sys_name': 2,
            'sys_ping_addr': 3,
            'comp_mgmt_url': 4,
            'comp_name': 5,
            'comp_ping_addr': 6,
            'sys_unique_id': 7
        }}),
        # TODO: edge case 'sys_unique_id' (GUID) needs conversion between string and GUID
        # For now we assume the blob is always a string
        ('blob', BytearrayField, None, {'hex': False}),
    ]



###########################################################################
#      ____  _____________  __    __  __ _           _____ ___   _        #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)  #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \      #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\     #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B     #
#                                                                         #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.          #
#                  SPDX-License-Identifier: BSD-3-Clause                  #
#                                                                         #
###########################################################################

import argparse
import os
import sys
import yaml
from datetime import datetime
import logging

from frugy.__init__ import __version__
from frugy.fru import Fru
from frugy.fru_registry import FruRecordType, rec_enumerate, rec_lookup_by_name, rec_info, schema_entry_info
from frugy.types import FruAreaChecksummed
from frugy.multirecords import MultirecordEntry


def list_supported_records():
    width = 108
    separator = '+' + '-' * (width-2) + '+'
    lf = f"|".ljust(width-1) + '|'
    print(separator)
    for rec_type in list(FruRecordType):
        rec_list = rec_enumerate(rec_type)
        if len(rec_list) != 0:
            print(lf)
            print(f'| type: {rec_type.name}'.ljust(width-1) + '|')
            print(lf)
            for r in rec_list:
                name, doc = rec_info(r)
                print(f'| {name.ljust(33)} {doc}'.ljust(width-1) + '|')
            print(lf)
            print(separator)


def list_record_schema(rec_name):
    schema = rec_lookup_by_name(rec_name)._schema
    print(f'{"Name".ljust(20)} {"Type".ljust(30)} {"Opt"}')

    for entry in schema:
        e_name, e_type, e_opt = schema_entry_info(entry)
        if e_name.startswith('_'):
            continue
        if e_opt is not None and 'constants' in e_opt:
            e_opt = ', '.join(k for k in e_opt['constants'].keys())
        else:
            e_opt = ''
        print(f'{e_name.ljust(20)} {e_type.ljust(30)} {e_opt}')


def writer(fname, content, bin_mode=False):
    if fname != '-':
        with open(fname, 'wb' if bin_mode else 'w') as f:
            f.write(content)
    else:
        if bin_mode:
            os.write(sys.stdout.fileno(), content)
        else:
            sys.stdout.write(content)


def dict_set(d, keys, item):
    if len(keys) > 1:
        key, rest = keys[0], keys[1:]
        if key not in d:
            d[key] = {}
        dict_set(d[key], rest, item)
    else:
        d[keys[0]] = item


def main():
    parser = argparse.ArgumentParser(
        description='FRU Generator YAML'
    )
    parser.add_argument('srcfile',
                        type=str,
                        nargs='?',
                        help='Source file for reading'
                        )
    parser.add_argument('--version',
                        action='version',
                        version='%(prog)s ' + __version__
                        )
    parser.add_argument('-o', '--output',
                        type=str,
                        help='output file (derived from input file if not set)'
                        )
    parser.add_argument('-w', '--write',
                        action='store_true',
                        help='FRU write mode (convert YAML to FRU image), default'
                        )
    parser.add_argument('-r', '--read',
                        action='store_true',
                        help='FRU read mode (convert FRU image to YAML)'
                        )
    parser.add_argument('-d', '--dump',
                        action='store_true',
                        help='dump FRU information to stdout (same as -r -o -)'
                        )
    parser.add_argument('-e', '--eeprom-size',
                        type=int,
                        help='pad FRU image to match EEPROM size in bytes (only valid in write mode)'
                        )
    parser.add_argument('-s', '--set',
                        type=str,
                        action='append',
                        help='set FRU record field to a value (only valid in write mode)'
                        )
    parser.add_argument('-t', '--timestamp',
                        action='store_true',
                        help='set BoardInfo.mfg_date_time timestamp to current UTC time (only valid in write mode)'
                        )
    parser.add_argument('-b', '--broken',
                        action='store_true',
                        help='enable workaround to parse Opal Kelly EEPROMs'
                        )
    parser.add_argument('-c', '--ignore-checksum-errors',
                        action='store_true',
                        help='ignore checksum errors when parsing a FRU image'
                        )
    parser.add_argument('-l', '--list',
                        type=str,
                        default=None,
                        const='',
                        nargs='?',
                        help='list supported FRU records or schema of specified record'
                        )
    parser.add_argument('-v', '--verbosity',
                        type=int,
                        help='set verbosity (0=quiet, 1=info, 2=debug)'
                        )
    args = parser.parse_args()

    verbosity = args.verbosity if args.verbosity is not None else 0
    logging.basicConfig(
        format='%(levelname)7s: %(message)s',
        level=[logging.WARNING, logging.INFO, logging.DEBUG][verbosity]
    )

    if args.list is not None:
        if args.list == '':
            list_supported_records()
        else:
            list_record_schema(args.list)
        sys.exit(0)

    if args.srcfile is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    read_mode = args.read or args.dump
    if args.write and args.read:
        parser.print_help(sys.stderr)
        sys.exit(1)

    if read_mode and (args.eeprom_size is not None or args.set or args.timestamp):
        parser.print_help(sys.stderr)
        sys.exit(1)

    if read_mode and args.broken:
        MultirecordEntry.opalkelly_workaround_enabled = True

    if read_mode and args.ignore_checksum_errors:
        FruAreaChecksummed.ignore_checksum_errors = True

    outfile = args.output
    if args.dump:
        if outfile:
            parser.print_help(sys.stderr)
            sys.exit(1)
        outfile = '-'

    basename, ext = os.path.splitext(os.path.basename(args.srcfile))

    if not read_mode and ext != '.yml' and ext != '.yaml':
        print('Cowardly refusing to read a YAML file not ending with .yaml or .yml', file=sys.stderr)
        sys.exit(1)

    if not outfile:
        basename, _ = os.path.splitext(os.path.basename(args.srcfile))
        outfile = basename + ('.yml' if read_mode else '.bin')

    fru = Fru()

    if read_mode:
        try:
            fru.load_bin(args.srcfile)
            writer(outfile, fru.dump_yaml())
        except RuntimeError as e:
            print(f'Error while parsing or writing: {e}')
            return False
    else:
        with open(args.srcfile, 'r') as infile:
            fru_dict = yaml.safe_load(infile)

        if args.set is not None:
            for s in args.set:
                k, v = s.split('=')
                key_path = k.split('.')
                if len(key_path) > 1:
                    # Traverse hierarchy of selected key_path e.g. ['BoardInfo', 'serial_number']
                    dict_set(fru_dict, key_path, v)
                else:
                    # Set property e.g 'serial_number' of all first level records (e.g. 'BoardInfo', 'ProductInfo')
                    key = key_path[0]
                    for k in fru_dict.keys():
                        if key in fru_dict[k]:
                            fru_dict[k][key] = v

        if args.timestamp:
            if 'BoardInfo' in fru_dict:
                fru_dict['BoardInfo']['mfg_date_time'] = datetime.utcnow()
            else:
                print(
                    'Error: FRU needs BoardInfo area to carry the timestamp', file=sys.stderr)
                sys.exit(1)

        fru.update(fru_dict)
        img = fru.serialize()
        if args.eeprom_size is not None:
            if len(img) <= args.eeprom_size:
                img += b'\xff' * (args.eeprom_size - len(img))
            else:
                print(
                    f'Error: Image size ({len(img)}) exceeds EEPROM size ({args.eeprom_size})', file=sys.stderr)
                sys.exit(1)
        writer(outfile, img, bin_mode=True)


if __name__ == '__main__':
    main()



###########################################################################
#      ____  _____________  __    __  __ _           _____ ___   _        #
#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)  #
#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \      #
#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\     #
#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B     #
#                                                                         #
#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.          #
#                  SPDX-License-Identifier: BSD-3-Clause                  #
#                                                                         #
###########################################################################

from frugy.types import FruAreaBase, FixedField, FixedStringField, GuidField, ArrayField, BytearrayField, IpV4Field, bin2hex_helper
import bitstruct
from frugy.fru_registry import FruRecordType, rec_register, rec_lookup_by_id, rec_lookup_by_name
from frugy.areas import ipmi_area
import logging
import frugy.fru


@ipmi_area
class MultirecordArea:
    ''' Platform Management FRU Information Storage Definition, Table 16-1 '''
    ''' This is a daisy-chain of MultirecordEntry objects '''

    def __init__(self, initdict=None):
        self.records = []
        if initdict is not None:
            self.update(initdict)

    def update(self, initdict):
        self.records = []
        for v in initdict:
            try:
                constructor = rec_lookup_by_name(v['type'])
            except KeyError:
                raise RuntimeError(f"Unknown multirecord entry {v['type']}")
            self.records.append(constructor(v))

    def __repr__(self):
        return self.to_dict().__repr__()

    def to_dict(self):
        return [v.to_dict() for v in self.records]

    def serialize(self):
        result = b''
        for i, v in enumerate(self.records):
            v.end_of_list = 1 if i == len(self.records)-1 else 0
            result += v.serialize()
        return result

    def deserialize(self, input):
        self.records = []
        remainder = input
        while len(remainder):
            new_entry, remainder, end_of_list = MultirecordEntry.deserialize(
                remainder)
            if new_entry is not None:
                self.records.append(new_entry)
            if end_of_list:
                break

        return remainder

    def size_total(self):
        return sum([v.size_total() for v in self.records])


class MultirecordEntry(FruAreaBase):
    _format_version = 2
    _multirecord_header_fmt = 'u8u1u3u4u8u8'

    opalkelly_workaround_enabled = False

    def __init__(self, initdict=None):
        self.end_of_list = 0
        if initdict is not None:
            # for MultirecordEntry, type is used for type identification, not for the fields
            initdict.pop('type', None)
        super().__init__(initdict=initdict)

    def size_payload(self):
        # Add header size
        return super().size_payload() + 5 + len(self._payload_prologue())

    def _payload_prologue(self):
        # data prepended before actual payload by subclasses
        return b''

    def to_dict(self):
        # save type identification
        result = {'type': self.__class__.__name__}
        result.update(super().to_dict())
        return result

    def serialize(self):
        payload = self._payload_prologue() + super().serialize()
        payload_cksum = (-sum(payload)) & 0xff
        header = bitstruct.pack(self._multirecord_header_fmt,
                                self._type_id,
                                self.end_of_list,
                                0,
                                self._format_version,
                                len(payload),
                                payload_cksum)
        header_cksum = (-sum(header)) & 0xff
        header += header_cksum.to_bytes(length=1, byteorder='little')
        return header + payload

    @classmethod
    def deserialize(cls, input):
        header_fmt_complt = MultirecordEntry._multirecord_header_fmt + 'u8'
        header_len = bitstruct.calcsize(header_fmt_complt) // 8
        type_id, end_of_list, _, format_version, \
            payload_len, payload_cksum, _ = bitstruct.unpack(
                header_fmt_complt, input)
        header, remainder = input[:header_len], input[header_len:]
        payload, remainder = remainder[:payload_len], remainder[payload_len:]

        logging.debug(
            f"{cls.__name__}: Trying to deserialize multirecord type_id=0x{type_id:02x}, len={len(header)+len(payload)}")
        logging.debug(
            f"{cls.__name__}: header: {bin2hex_helper(header)}, payload: {bin2hex_helper(payload)}")

        try:
            if sum(header) & 0xff != 0:
                end_of_list = 1
                raise RuntimeError("MultirecordEntry header checksum invalid")

            if cls.opalkelly_workaround_enabled and len(payload) == 0:
                # Opal Kelly seems to mark the end of list with an empty payload multirecord
                end_of_list = 1
                return None, remainder, end_of_list

            if (sum(payload) + payload_cksum) & 0xff != 0:
                end_of_list = 1
                raise RuntimeError("MultirecordEntry payload checksum invalid")

            try:
                cls_id = rec_lookup_by_id(
                    FruRecordType.ipmi_multirecord, type_id)
            except KeyError:
                raise RuntimeError(f"Unknown multirecord type 0x{type_id:02x}")

            if hasattr(cls_id, 'from_payload'):
                new_entry = cls_id.from_payload(payload)
            else:
                new_entry = cls_id()
                new_entry._deserialize(payload)

            new_entry._type_id = type_id
            new_entry._format_version = format_version
            new_entry.end_of_list = end_of_list

        except RuntimeError as e:
            logging.warning(f"Failed to deserialize multirecord, type_id=0x{type_id:02x}, end_of_list={end_of_list}, "
                            f"format_version={format_version}, len={len(header)+len(payload)}")
            logging.warning(f"reason: {e}")
            logging.warning(
                f"header: {bin2hex_helper(header)}, payload: {bin2hex_helper(payload)}")
            frugy.fru.import_log(
                f'Failed to deserialize multirecord 0x{type_id:02x} ({e})')
            new_entry = None

        except ValueError as e:
            # Vendor ID mismatch: Don't issue a warning, just ignore it
            logging.debug(f"{e}")
            logging.debug(
                f"Silently ignoring private / proprietary multirecord")
            frugy.fru.import_log(
                f'Ignored private / proprietary multirecord ({e})')
            new_entry = None

        except EOFError as e:
            # Empty payload: Issue warning, but try to proceed
            logging.warning(f"{e}")
            new_entry = None

        return new_entry, remainder, end_of_list


def ipmi_multirecord(rec_id):
    def register_and_set_id(cls):
        cls._type_id = rec_id
        rec_register(cls, FruRecordType.ipmi_multirecord, rec_id)
        return cls
    return register_and_set_id
